/*PCI Express protocol decoder.
*******************************

Gavin Lofts

Description
This module attaches to an unscrambled PIPE interface.  Future work would be derializer and descrambler.
3 output files are generated: .log contains all data (packets and ordered sets). .log_tlp
***********************************/

//Current Limitations
//No decoding for the EQ TS1 and EQ TS2 ordered sets.  Only standard TS1 and TS2 ordered sets are supported
//Needs to be updated to handle END symbols in lane 3 of x8 links

//TLP formats
`define DW3_NODAT 3'h0
`define DW4_NODAT 3'h1
`define DW3_DAT   3'h2
`define DW4_DAT   3'h3
`define TLP_PRE   3'h4

//TLP types
`define MRD  5'h00
`define MRDLK 5'h01
`define MWR 5'h0
`define IORD 5'h02
`define IOWR 5'h02
`define CFGRD0 5'h04
`define CFGWR0 5'h04
`define CFGRD1 5'h05
`define CFGWR1 5'h05
`define CPL 5'h0a
`define CPLD 5'h0a
`define CPLLK 5'h0b
`define CPLDLK 5'h0b
`define FETCHADD 5'h0c
`define SWAP 5'h0d
`define CAS 5'h0e

//K codes
`define PAD 8'hf7
`define COM 8'hbc
`define STP 8'hfb
`define SDP 8'h5c
`define END 8'hfd
`define EDB 8'hfe
`define PAD 8'hf7
`define SKP 8'h1c
`define FTS 8'h3c
`define IDL 8'h7c
`define EIE 8'hfc

//DLLP types
`define ACK 8'h00
`define NAK 8'h10
`define PM_DLLP 8'h20
`define INIT_FC1_P 8'h40
`define INIT_FC1_NP 8'h50
`define INIT_FC1_CPL 8'h60
`define INIT_FC2_P 8'hC0
`define INIT_FC2_NP 8'hD0
`define INIT_FC2_CPL 8'hE0
`define UPDATE_FC_P 8'h80
`define UPDATE_FC_NP 8'h90
`define UPDATE_FC_CPL 8'hA0

module pcie_proto_dec (
	input wire		reset,
	input wire		clk,
	input wire [7:0]	rxdata0,
	input wire 		rxdatak0,
	input wire		rx0_en,
	input wire [7:0]	rxdata1,
	input wire 		rxdatak1,
	input wire		rx1_en,
	input wire [7:0]	rxdata2,
	input wire 		rxdatak2,
	input wire		rx2_en,
	input wire [7:0]	rxdata3,
	input wire 		rxdatak3,
	input wire		rx3_en,
	input wire [7:0]	rxdata4,
	input wire 		rxdatak4,
	input wire		rx4_en,
	input wire [7:0]	rxdata5,
	input wire 		rxdatak5,
	input wire		rx5_en,
	input wire [7:0]	rxdata6,
	input wire 		rxdatak6,
	input wire		rx6_en,
	input wire [7:0]	rxdata7,
	input wire 		rxdatak7,
	input wire		rx7_en
	);

typedef enum int
{ 	IDLE,
	ORDERED_SET,
	TSx_LINK,
	TSx_FTS,
	TSx_CONTR,
	TSx_RATE,
	TSx_IDENT,
	TS1_COUNT,
	TS2_COUNT,
	FTS,
	EIOS,
	EIEOS,
	SKIP,
	TLP,
	DLLP
} rx_state_t;

rx_state_t rx_state;
	
parameter START_TEXT = "PCIe Receive decode:";
parameter LOGFILE_NAME = "pcie_logfile.log";
parameter LANE_ASSIGNMENT_AUTO = 1'b1;
parameter LANE_ASSIGNMENT_INIT0 = -1;
parameter LANE_ASSIGNMENT_INIT1 = -1;
parameter LANE_ASSIGNMENT_INIT2 = -1;
parameter LANE_ASSIGNMENT_INIT3 = -1;
parameter LANE_ASSIGNMENT_INIT4 = -1;
parameter LANE_ASSIGNMENT_INIT5 = -1;
parameter LANE_ASSIGNMENT_INIT6 = -1;
parameter LANE_ASSIGNMENT_INIT7 = -1;
integer LOGFILE;
integer LOGFILE_TLP;
integer LOGFILE_DLLP;
parameter NUM_LANES_INIT = 8;

integer i;
integer j;

reg	[7:0]	rxdata_int[7:0];
reg       rxdatak_int[7:0];
reg       rx_en_int[7:0];

reg [7:0] enable_mask;

integer num_lanes;
integer error_flag;
integer num_skips;

string message;
string filename;

//Training sequence params
string link_number;
string lane_number[7:0];
string lane_number_con;
string n_fts;
reg [7:0]   training_control;
string data_rate;
integer ts1_counter;
integer ts2_counter;

//Array which contains lane assignments
integer lane_assignment_array[7:0];
reg lane_assignment_done;
integer lane0_index;

//Packet buffer (For TLP and DLLP)
reg [7:0] packet_buf[4095:0];
integer packet_pointer;
integer lane_pointer;

//Initialize variables from paramters
initial
begin
	num_lanes = NUM_LANES_INIT;
	lane_assignment_done = !(LANE_ASSIGNMENT_AUTO);
	if(LANE_ASSIGNMENT_AUTO == 1'b0)
	  begin
	    lane_assignment_array[0] = LANE_ASSIGNMENT_INIT0;
	    lane_assignment_array[1] = LANE_ASSIGNMENT_INIT1;
	    lane_assignment_array[2] = LANE_ASSIGNMENT_INIT2;
	    lane_assignment_array[3] = LANE_ASSIGNMENT_INIT3;
	    lane_assignment_array[4] = LANE_ASSIGNMENT_INIT4;
	    lane_assignment_array[5] = LANE_ASSIGNMENT_INIT5;
	    lane_assignment_array[6] = LANE_ASSIGNMENT_INIT6;
	    lane_assignment_array[7] = LANE_ASSIGNMENT_INIT7;
	  end
	else
	  begin
	    for(i = 0; i < 8; ++i)
	    begin
	      lane_assignment_array[i] = -1;
	    end
	  end
end

initial
begin
	rxdata_int[0] = 8'hFF;
	rxdatak_int[0] = 1'b0;
	LOGFILE = $fopen(LOGFILE_NAME);
	$sformat(filename, "%s%s", LOGFILE_NAME, "_tlp");
	LOGFILE_TLP = $fopen (filename);
	$sformat(filename, "%s%s", LOGFILE_NAME, "_dllp");
	LOGFILE_DLLP = $fopen (filename);	
	$fwrite(LOGFILE, "PCIe decoded output\n");
	$sformat(message, "%s %s %s", "filename:", LOGFILE_NAME, "\n\n");
	$fwrite(LOGFILE, message);
end
//Resolve z to 0
buf buf7 (rx_en_int[7], rx7_en);
buf buf6 (rx_en_int[6], rx6_en);
buf buf5 (rx_en_int[5], rx5_en);
buf buf4 (rx_en_int[4], rx4_en);
buf buf3 (rx_en_int[3], rx3_en);
buf buf2 (rx_en_int[2], rx2_en);
buf buf1 (rx_en_int[1], rx1_en);
buf buf0 (rx_en_int[0], rx0_en);
pulldown pull7 (rx7_en);
pulldown pull6 (rx6_en);
pulldown pull5 (rx5_en);
pulldown pull4 (rx4_en);
pulldown pull3 (rx3_en);
pulldown pull2 (rx2_en);
pulldown pull_1 (rx1_en);
pulldown pull_0 (rx0_en);

	
//physical layer framing
always @(posedge clk or posedge reset)
begin
	if (reset == 1'b1)
	begin
		rxdata_int[0] <= 8'hFF;
	end
	else
	begin
		rxdata_int[0] <= rxdata0;
		rxdatak_int[0] <= rxdatak0;
		rxdata_int[1] <= rxdata1;
		rxdatak_int[1] <= rxdatak1;
		rxdata_int[2] <= rxdata2;
		rxdatak_int[2] <= rxdatak2;
		rxdata_int[3] <= rxdata3;
		rxdatak_int[3] <= rxdatak3;
		rxdata_int[4] <= rxdata4;
		rxdatak_int[4] <= rxdatak4;
		rxdata_int[5] <= rxdata5;
		rxdatak_int[5] <= rxdatak5;			
		rxdata_int[6] <= rxdata6;
		rxdatak_int[6] <= rxdatak6;
		rxdata_int[7] <= rxdata7;
		rxdatak_int[7] <= rxdatak7;
			
		case (rx_state)
		  
		  DLLP:
		    begin
		      //Assign lanes to the packet buffer in the correct order
		       for(i = 0; i < 8; ++i)
		       begin
		        if(lane_assignment_array[i] != -1)
		          begin
		            packet_buf[(packet_pointer + lane_assignment_array[i])] = rxdata_int[lane_assignment_array[i]];
		            lane_pointer++;
		            if(rxdatak_int[i])
		              begin
		                if(i == (num_lanes - 1))
		                  begin
		                    if (rxdata_int[lane_assignment_array[i]] ==`END)
		                      begin
		                        //$sformat(message, "DLLP received at time %10d\n", $time);
		                       // $fwrite (LOGFILE, message);
		                        packet_pointer = packet_pointer + lane_pointer; /* ensure the pointer is up to date before calling the decoder*/
		                        j = packet_decode_dllp(packet_buf, packet_pointer, $time);
		                        rx_state = IDLE;
		                      end
		                    else
		                      begin
		                            $sformat(message, "!!!ERROR!!!, invalid K symbol found in DLLP at time %10d\n received  %02x, expected END\n", rxdata_int[i], $time);
		                            $fwrite (LOGFILE, message);
		                            rx_state = IDLE;
		                      end
		                  end
		                else
		                  begin
		                   $sformat(message, "!!!ERROR!!!,K symbol %02x found in DLLP in incorrect lane at time %10d\n\n", rxdata_int[i], $time); 
		                   $fwrite (LOGFILE, message);
		                   rx_state = IDLE;
		                  end
		              end
		          end
		       end 	
		       packet_pointer = packet_pointer + lane_pointer;	  
		       lane_pointer = 0;   
		    end		  
		  
		  TLP:
		    begin
		      //Assign lanes to the packet buffer in the correct order
		       for(i = 0; i < 8; ++i)
		       begin
		        if(lane_assignment_array[i] != -1)
		          begin
		            packet_buf[(packet_pointer + lane_assignment_array[i])] = rxdata_int[lane_assignment_array[i]];
		            lane_pointer++;
		            if(rxdatak_int[i])
		              begin
		                if(i == (num_lanes - 1))
		                  begin
		                    if (rxdata_int[lane_assignment_array[i]] ==`END)
		                      begin
		                        //$sformat(message, "TLP received at time %10d\n", $time);
		                        //$fwrite (LOGFILE, message);
		                        packet_pointer = packet_pointer + lane_pointer; /* ensure the pointer is up to date before calling the decoder*/		                        
		                        j = packet_decode(packet_buf, packet_pointer, $time);
		                        rx_state = IDLE;
		                      end
		                    else
		                      begin
		                        if (rxdata_int[lane_assignment_array[i]] ==`EDB)
		                          begin
		                            $sformat(message, "TLP,Nullified TLP received at time %10d\n", $time);
		                            $fwrite (LOGFILE, message);
		                            rx_state = IDLE;
		                          end
		                        else
		                          begin
		                            $sformat(message, "!!!ERROR!!!, invalid K symbol %02x found in TLP at time %10d\n\n", rxdata_int[i], $time);
		                            $fwrite (LOGFILE, message);
		                            rx_state = IDLE;
		                          end
		                      end
		                  end
		                else
		                  begin
		                   $sformat(message, "!!!ERROR!!!, invalid K symbol %02x found in TLP at time %10d\n\n", rxdata_int[i], $time); 
		                   $fwrite (LOGFILE, message);
		                   rx_state = IDLE;
		                  end
		              end
		          end
		       end 	
		       packet_pointer = packet_pointer + lane_pointer;	  
		       lane_pointer = 0;   
		    end
		  
		  
		  TS1_COUNT:
		    begin
		      //Check lanes are identical
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))
		        begin
		          if(rxdata_int[lane0_index] == 8'h4A)
		            begin
		              if(ts1_counter > 8)
		                begin
		                  //$sformat(message, "TS1 ordered set received at time %10d\n", $time);
		                  //$fwrite(LOGFILE, message);
		                  $sformat(message, "Ordered Set,TS1,Link Number,Lane Number,N_FTS,Data Rate,Training Control,time\n");
		                  $fwrite(LOGFILE, message);
		                  $sformat(message, "Ordered Set,TS1,%s,%s,%s,%s,'%08b,%10d\n\n", link_number, lane_number_con, n_fts, data_rate, training_control,$time);
		                  $fwrite(LOGFILE, message);
		                  rx_state = IDLE;
		                end
		              else
		                begin
		                  ts1_counter++;
		                end
		            end
		          else
		            begin
		              rx_state = IDLE;
		              $sformat(message, "!!!ERROR!!!,Error in TS1 training sequence at time %10d, char received %2x\n\n", $time, rxdata_int[0]);
		              $fwrite(LOGFILE, message);
		            end
		        end
		      else
		        begin
		          rx_state = IDLE;
		        end
		    end
		    
		   TS2_COUNT:
		    begin
		      //Check lanes are identical
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))	      
		        begin
		          if(rxdata_int[lane0_index] == 8'h45)
		            begin
		              if(ts1_counter > 8)
		                begin
		                  //$sformat(message, "TS2 ordered set received at time %10d\n", $time);
		                  //$fwrite(LOGFILE, message);
		                  $sformat(message, "Ordered Set,TS2,Link Number,Lane Number,N_FTS,Data Rate,Training Control,time\n");
		                  $fwrite(LOGFILE, message);
		                  $sformat(message, "Ordered Set,TS2,%s,%s,%s,%s,'%08b,%10d\n\n", link_number, lane_number_con, n_fts, data_rate, training_control,$time);
		                  $fwrite(LOGFILE, message);
		                  rx_state = IDLE;
		                end
		              else
		                begin
		                  ts1_counter++;
		                end
		            end
		          else
		            begin
		              rx_state = IDLE;
		              $sformat(message, "!!!ERROR!!!,Error in TS2 training sequence at time %10d, char received %2x\n\n", $time, rxdata_int[0]);
		              $fwrite(LOGFILE, message);
		            end
		        end
		      else
		        begin
		          rx_state = IDLE;
		        end
		    end
		    
		  TSx_IDENT:
		    begin
		      //Check lanes are identical
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))
				    begin
				      if(rxdatak_int[lane0_index] == 1'b0)
				        begin
				          if(rxdata_int[lane0_index] == 8'h4A)
				            begin
				              rx_state = TS1_COUNT;
				              ts1_counter = 1;
				            end
				          else
				            begin
				              if(rxdata_int[lane0_index] == 8'h45)
				                begin
				                  rx_state = TS2_COUNT;
				                  ts2_counter = 1;
				                end
				              else
				                begin
				                  $sformat(message, "!!!ERROR!!!,Error invalid training sequence code at char 6 = %02x at time %10d\n\n", rxdata_int[0], $time);
				                  $fwrite(LOGFILE, message);
				                  rx_state = IDLE;
				                end
				            end
				        end
				      else
				        begin
				          $sformat(message, "!!!ERROR!!!,Error training sequence with K code at char 6 at time %10d\n\n", $time);
				          $fwrite(LOGFILE, message);
				        end
				    end
				  else
				    begin
				      rx_state = IDLE;
				    end
		    end
		  TSx_RATE:
		  begin
//Check that all lanes match during an ordered set.  Mandadtory for rate bits
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))
				  begin
            if(rxdata_int[lane0_index] & 8'h02)
              begin
                $sformat(data_rate, "2.5GT/s");
                if(rxdata_int[lane0_index] & 8'h04)
                  begin
                    $sformat(data_rate, "%s 5.0GT/s", data_rate);
                    if(rxdata_int[lane0_index] & 8'h08)
                      begin
                        $sformat(data_rate, "%s 8.0GT/s", data_rate);
                      end
                  end
                if(rxdata_int[lane0_index] & 8'h40)
                  begin
                    $sformat(data_rate, "%s bit 6 set", data_rate);
                  end
                if(rxdata_int[lane0_index] & 8'h80)
                  begin
                    $sformat(data_rate, "%s speed_change", data_rate);
                  end
                //$fwrite(LOGFILE, data_rate);
                rx_state = TSx_CONTR;
              end
            else
              begin
                $sformat(message, "!!!ERROR!!!, 2.5GT/s must be supported.  Data rate identifier: %02x, time %10d\n\n", rxdata_int[0], $time);
                $fwrite(LOGFILE, message);
                rx_state = IDLE;
              end
				  end
				else
				  begin
				    rx_state = IDLE;
				  end
		  end
		  TSx_CONTR:
		  begin
//Check that all lanes match during an ordered set.  Mandadtory for control bits
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))
				  begin
				    training_control = rxdata_int[lane0_index];
				    //$sformat(message, "training control bits: %01b %01b %01b %01b %01b %01b %01b %01b\n", training_control[0], training_control[1], training_control[2],
				    //training_control[3], training_control[4], training_control[5], training_control[6], training_control[7]);
				    //$fwrite(LOGFILE, message);
				    rx_state = TSx_IDENT;
				  end
				else
				  begin
				    rx_state = IDLE;
				  end
		  end
//In these states the the ordered set is known to be a training sequence, but we don't know if it is TS1 or TS2		
      TSx_FTS:
      begin
//Check that all lanes match during an ordered set.  Mandadtory for N_FTS
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))
				  begin
				    $sformat(n_fts, "%03d", rxdata_int[lane0_index]);
				    rx_state = TSx_RATE;
				  end
				else
				  begin
				    //Print made by comapre
				    rx_state = IDLE;
				  end 
      end
      
        
		  TSx_LINK:
		  begin
		    rx_state = IDLE;
		    error_flag = 0;
		    
		    for (i = 0; i < 8; ++i)
		    begin
		      if(rx_en_int[i] == 1'b1)
		        begin
		          if(rxdatak_int[i] == 1'b1)
		            begin
		              if(rxdata_int[i] == `PAD)
		                begin
		                  $sformat(lane_number[i], "%02d = PAD", i);
		                  lane_assignment_array[i] = -1;
		                end
		              else
		                begin
		                  ++error_flag;
		                end
		            end
		          else
		            begin
		              $sformat(lane_number[i], "%02d = %02d", i, rxdata_int[i]);
		              lane_assignment_array[i] = rxdata_int[i];
		              num_lanes = (rxdata_int[i] + 1'b1);
		              lane_assignment_done = 1'b1;
		            end
		        end
		    end
		    if(error_flag > 0)
		      begin
		        $sformat(message, "!!!ERROR!!!,Invalid character in link number field of training sequence at time %d\n\n", $time);
		        $fwrite(LOGFILE, message);
		        rx_state = IDLE;
		      end    
		    else
		      begin
		        $sformat(lane_number_con, "%s %s %s %s %s %s %s %s", lane_number[0], lane_number[1], lane_number[2], lane_number[3], 
		        lane_number[4], lane_number[5], lane_number[6], lane_number[7]);
		        //$fwrite(LOGFILE, "\n\nsetting lane...");
		        //$fwrite(LOGFILE, lane_number_con);
		        rx_state = TSx_FTS;
		      end      		      		      		      		      		      		      
		  end
		  
//Structured like a skip, as the ordered set is the same just with the FTS character instead of SKP		  
		  FTS:
		  begin
		    if (rxdatak_int[lane0_index] == 1'b1 && rxdata_int[lane0_index] == `FTS)
		      begin
		        error_flag = 0;
		        num_skips++;
//Check that all lanes match during an ordered set.  Mandadtory for Skip sets and the first two chars of TSx
				   	if(check_identical(rx_en_int, rxdatak_int, rxdata_int))
				      begin
		          if (num_skips > 2)
		            begin
		              $sformat(message, "FTS,time\n", $time);
		              $fwrite(LOGFILE, message);
		              $sformat(message, "FTS,%10d\n\n", $time);
		              $fwrite(LOGFILE, message);				              
		              rx_state = IDLE;
		            end
		          end 
		         else
		          begin
		            //Print performed by check function
		            rx_state = IDLE;
		          end         		          		          		          
		      end
		    else
		      begin
		        $sformat(message, "!!!ERROR!!!,During FTS sequence, FTS character not received at time %10d\n\n", $time);
		        $fwrite(LOGFILE, message);
		      end

		  end  
		  
//Structured like a skip, as the ordered set is the same just with the IDL character instead of SKP		  
		  EIOS:
		  begin
		    if (rxdatak_int[lane0_index] == 1'b1 && rxdata_int[lane0_index] == `IDL)
		      begin
		        error_flag = 0;
		        num_skips++;
//Check that all lanes match during an ordered set.  Mandadtory for Skip sets and the first two chars of TSx
				   	 if(check_identical(rx_en_int, rxdatak_int, rxdata_int))	  		          		        		          		
				      begin
		          if (num_skips > 2)
		            begin
		              $sformat(message, "EIOS,time\n", $time);
		              $fwrite(LOGFILE, message);
		              $sformat(message, "EIOS,%10d\n\n", $time);
		              $fwrite(LOGFILE, message);		
		              rx_state = IDLE;
		            end
		          end 
		         else
		          begin
		            //Print performed by check function
		            rx_state = IDLE;
		          end         		          		          		          
		      end
		    else
		      begin
		        $sformat(message, "!!!ERROR!!! During EIOS sequence, EIOS character not received at time %10d\n\n", $time);
		        $fwrite(LOGFILE, message);
		      end

		  end 		  
		  
//Structured like a skip, as the ordered set is the same just with the IDL character instead of SKP, and 14 characters	  
		  EIEOS:
		  begin
		    if (rxdatak_int[lane0_index] == 1'b1 && rxdata_int[lane0_index] == `EIE)
		      begin
		        error_flag = 0;
		        num_skips++;
//Check that all lanes match during an ordered set.  Mandadtory for Skip sets and the first two chars of TSx
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))  		          		        		          		
				      begin
		          end 
		         else
		          begin
		            //Print performed by check function
		            rx_state = IDLE;
		          end         		          		          		          
		      end
		    else
		      begin
		        if (rxdatak_int[lane0_index] == 1'b0 && rxdata_int[lane0_index] == 8'h4A && num_skips == 14)
		          begin
		            $sformat(message, "EIEOS,time\n", $time);
		            $fwrite(LOGFILE, message);
		            $sformat(message, "EIEOS,%10d\n\n", $time);
		            $fwrite(LOGFILE, message);		            
		          end
		        else
		          begin
		            $sformat(message, "!!!ERROR!!!,During EIEOS sequence, EIEOS character not received at time %10d\n\n", $time);
		            $fwrite(LOGFILE, message);
		          end
		      end

		  end 		  
		  
		  		  
//In this state the ordered set is Skip.  This is always COM + 3xSKP from transmitters.		  
		  SKIP:
		  begin
		    if (rxdatak_int[lane0_index] == 1'b1 && rxdata_int[lane0_index] == `SKP)
		      begin
		        error_flag = 0;
		        num_skips++;
//Check that all lanes match during an ordered set.  Mandadtory for Skip sets and the first two chars of TSx
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int)) 		          		        		          		
				      begin
		          if (num_skips > 2)
		            begin
		              $sformat(message, "Skip,Number of SKP sets,time\n");
		              $fwrite(LOGFILE, message);
		              $sformat(message, "Skip,%4d,%10d\n\n", num_skips,$time);
		              $fwrite(LOGFILE, message);		              
		              rx_state = IDLE;
		            end
		          end 
		         else
		          begin
		            //Print performed by check function
		            rx_state = IDLE;
		          end         		          		          		          
		      end
		    else
		      begin
		        $sformat(message, "!!!ERROR!!!,During Skip sequence, skip character not received at time %10d\n\n", $time);
		        $fwrite(LOGFILE, message);
		      end

		  end
		  
//Ini this state we have recieved COM indicating an ordered set, but don't know which ordered set it is		  
			ORDERED_SET:
			begin
				num_lanes = rx_en_int[0] + rx_en_int[1] + rx_en_int[2] + rx_en_int[3] + rx_en_int[4] + rx_en_int[5] + rx_en_int[6] + rx_en_int[7];
				//$sformat(message, "Ordered set, number of lanes = %01d, time is %10d\n", num_lanes, $time);
				//$fwrite(LOGFILE, message);
//Check that all lanes match during an ordered set.  Mandadtory for Skip sets and the first two chars of TSx
		      if(check_identical(rx_en_int, rxdatak_int, rxdata_int))	
				begin
				 //Check for skip char (meaning Skip ordered set)
				if(rxdatak_int[lane0_index] == 1)
				  begin
				    if (rxdata_int[lane0_index] == `SKP)
				      begin
				        rx_state = SKIP;
				        error_flag = 0;
				        num_skips = 1;      
				      end	
				    else
				      begin
				        if (rxdata_int[lane0_index] == `PAD)
				          begin
				            rx_state = TSx_LINK;
				            $sformat(link_number, "PAD");
				          end
				        else
				          begin
				            if(rxdata_int[lane0_index] ==`FTS)
				              begin
				                rx_state = FTS;
				                error_flag = 0;
				                num_skips = 1;
				              end
				            else
				              begin
				                if(rxdata_int[lane0_index] ==`IDL)
				                  begin
				                    rx_state = EIOS;
				                    error_flag = 0;
				                    num_skips = 1;
				                  end
				                else
				                  begin
				                    if(rxdata_int[lane0_index] ==`EIE)
				                      begin
				                        rx_state = EIEOS;
				                        error_flag = 0;
				                        num_skips = 1;
				                      end
				                    else
				                      begin
				                        $sformat(message, "!!!ERROR!!!,invalid K code after comma character: %x\n", rxdata_int[0]);
				                        $fwrite(LOGFILE, message);
				                        rx_state = IDLE;
				                      end
				                  end
				              end
				          end
				      end
				  end
				else
				  begin
				    rx_state = TSx_LINK;
				    $sformat(link_number, "%03d", rxdata_int[lane0_index]);
				  end
				end
				else
				  begin
				  //Compare function returns message for lanes
				  rx_state = IDLE;
				  end
			end //Ordered set case
			
			
//Look for start of TLLP, DLLP or ordered set			
		  IDLE:
		  begin
//Lane reversal stuff	    
//Check that all of the valid lanes are active if lane assignment is complete
		    if (lane_assignment_done)
		      begin
		        for (i = 0 ; i < 8 ; ++i)
		        begin
		          if (lane_assignment_array[i] > -1)
		            begin
		              if (rx_en_int[i] == 1'b0)
		                begin
		                  $sformat(message, "lane %01d assigned but not active.  Returning to unassigned\n\n", i);
		                  $fwrite(LOGFILE, message);
		                  lane_assignment_done = 1'b0;
		                end
		            end
		        end
		      end
    //Which lane to check.  Should be lane 0 if lane assignment is complete.  lane0_index is assigned lane 0.  Otherwise it is just any lane with data
        if(lane_assignment_done)
          begin
            for (i = 0; i < 8; ++i)
            begin
              if(lane_assignment_array[i] == 0)
                begin
                  lane0_index = i;
                end
            end
          end
        else
          begin
            for(i = 0; i < 8; ++i)
            begin
              if(rx_en_int[i])
                begin
                  lane0_index = i;
                end
            end
          end
        if((rxdatak_int[lane0_index] == 1'b1) & (rx_en_int[lane0_index] ==1'b1))
		    begin
		      case(rxdata_int[lane0_index])
		        `COM:  //K28.5 COM Comma
		        begin
		          $sformat(message, "%s", "comma lane0,");
				      rx_state = ORDERED_SET;
				      error_flag = 0;
				      
				      for (i = 0; i < 8; ++i)
				      begin
				        if(rx_en_int[i] == 1'b1)
				          begin
				            if(rxdata_int[i]  != `COM || rxdatak_int[i] != 1'b1)
				              begin
				                error_flag = 1;
				              end
				          end
				      end			        				        				        				        				            				        				        				        				        				        				          
		  
				      if(error_flag == 1)
				        begin
					        $sformat(message, "%s", "\nERROR not all lanes contain comma\n");
				        end
				  
		        end //Case COM
		        
		        
		        `STP:  //K27.7 STP Start TLP
		        begin
		          if(lane_assignment_done == 1)
		            begin
		              /*packet_pointer keeps track of the whole packet, lane_pointer just counts how many valid lanes have been inserted*/
		              packet_pointer = 0;
		              lane_pointer = 0;
		              //Assign lanes to the packet buffer in the correct order
		              for(i = 0; i < 8; ++i)
		              begin
		                if(lane_assignment_array[i] != -1)
		                  begin
		                    packet_buf[(packet_pointer + lane_assignment_array[i])] = rxdata_int[lane_assignment_array[i]];
		                    lane_pointer++;
		                  end
		              end 
		              rx_state = TLP;
		              packet_pointer = packet_pointer + lane_pointer;
		              lane_pointer = 0;
		            end
		          else
		            begin
		              $sformat(message, "!!!ERROR!!! TLP received while lanes have not been assigned at time %10d\n\n", $time);
		              $fwrite(LOGFILE, message);
		              rx_state = IDLE;
		            end                		            		            		            		            		            
		        end		    //Case STP
		        
		            
		        `SDP:  //K28.2 SDP Start DLLP
		        begin
		          if(lane_assignment_done == 1)
		            begin
		              /*packet_pointer keeps track of the whole packet, lane_pointer just counts how many valid lanes have been inserted*/
		              packet_pointer = 0;
		              lane_pointer = 0;
		              //Assign lanes to the packet buffer in the correct order
		              for(i = 0; i < 8; ++i)
		              begin
		                if(lane_assignment_array[i] != -1)
		                  begin
		                    packet_buf[(packet_pointer + lane_assignment_array[i])] = rxdata_int[lane_assignment_array[i]];
		                    lane_pointer++;
		                  end
		              end 
		              rx_state = DLLP;
		              packet_pointer = packet_pointer + lane_pointer;
		              lane_pointer = 0;
		            end
		          else
		            begin
		              $sformat(message, "!!!ERROR!!!,DLLP received while lanes have not been assigned at time %10d\n\n", $time);
		              $fwrite(LOGFILE, message);
		              rx_state = IDLE;
		            end                		            		            		            		            		            
		        end		    //Case SDP
		        
		         
		        `END:  //K29.7 END End (DLLP or TLP)
		        begin
		          $sformat(message, "!!!ERROR!!!,Invalid END character received at time %10d\n\n", $time);
		          $fwrite(LOGFILE, message);
		        end		  //Case END
		              
		        `EDB:  //K30.7 EDB EnD Bad
		        begin
		          $sformat(message, "!!!ERROR!!!,Invalid EDB character received at time %10d\n\n", $time);
		          $fwrite(LOGFILE, message);
		        end		 //Case EDB
		               
		        `PAD:  //K23.7 PAD Pad
		        begin
		          $sformat(message, "!!!ERROR!!!,Invalid PAD character received at time %10d\n\n", $time);
		          $fwrite(LOGFILE, message);
		        end		 //Case PAD
		               
		        `SKP:  //K28.0 SKP Skip
		        begin
		          $sformat(message, "!!!ERROR!!!,Invalid SKP character received at time %10d\n\n", $time);
		          $fwrite(LOGFILE, message);
		        end		     //Case SKP
		           
		        `FTS:  //K28.1 FTS Fast Training Sequence
		        begin
		          $sformat(message, "!!!ERROR!!!,Invalid FTS character received at time %10d\n\n", $time);
		          $fwrite(LOGFILE, message);
		        end		     //Case FTS
		           
		        `IDL:  //K28.3 IDL Idle
		        begin
		          $sformat(message, "!!!ERROR!!!,Invalid IDL character received at time %10d\n\n", $time);
		          $fwrite(LOGFILE, message);
		        end		     //Case IDL
		           
		        `EIE:  //K28.7 EIE Electrical Idle Exit
		        begin
		          $sformat(message, "EIE,time\n");
		          $fwrite(LOGFILE, message);
		          $sformat(message, ",$10d\n\n", $time);
		          $fwrite(LOGFILE, message);
		        end		        //Case EIE
		        
		        default:  //invalid control character
		        begin
		          $sformat(message, "!!!ERROR!!!,Invalid control character %02x received at time %10d\n\n", $time, rxdata_int[0]);
		          $fwrite(LOGFILE, message);

		            rx_state <= IDLE;
		        end
		     endcase
		    end
		    else
		    begin
		     rx_state <= IDLE;
		    end	
		  end	
		  default:
		    rx_state <= IDLE;    
		  endcase
	 end
end

//Function to check that all lanes are identical (used for ordered sets)
function check_identical;
  input rx_en_int[7:0];         
  input rxdatak_int[7:0];
  input [7:0] rxdata_int[7:0];
  begin
        //$fwrite(LOGFILE, "call_func\n\n\n");
    				if ((rx_en_int[1] && ((rxdatak_int[0] == rxdatak_int[1]) && (rxdata_int[1] !=rxdata_int[1]))) || 
				(rx_en_int[2] && ((rxdatak_int[0] != rxdatak_int[2]) && (rxdata_int[2] !=rxdata_int[1]))) ||	
				(rx_en_int[3] && ((rxdatak_int[0] != rxdatak_int[3]) && (rxdata_int[3] !=rxdata_int[1]))) ||		
				(rx_en_int[4] && ((rxdatak_int[0] != rxdatak_int[4]) && (rxdata_int[4] !=rxdata_int[1]))) ||	
				(rx_en_int[5] && ((rxdatak_int[0] != rxdatak_int[5]) && (rxdata_int[5] !=rxdata_int[1]))) ||	
				(rx_en_int[6] && ((rxdatak_int[0] != rxdatak_int[6]) && (rxdata_int[6] !=rxdata_int[1]))) ||		
				(rx_en_int[7] && ((rxdatak_int[0] != rxdatak_int[7]) && (rxdata_int[7] !=rxdata_int[1]))))
				
				  begin
				    $sformat(message, "!!!Error!!! at time %d Not all characters match across lanes: lane0 %02x, k = %01b", $time, rxdata_int[0], rxdatak_int[0]);
				    if (rx_en_int[1])
				      begin
				       $sformat(message, "%s lane1 %02x k= %01b", message, rxdata_int[1], rxdatak_int[1]);
				      end
				    if (rx_en_int[2])
				      begin
				       $sformat(message, "%s lane1 %02x k= %01b", message, rxdata_int[2], rxdatak_int[2]);
				      end
				    if (rx_en_int[3])
				      begin
				       $sformat(message, "%s lane1 %02x k= %01b", message, rxdata_int[3], rxdatak_int[3]);
				      end
				    if (rx_en_int[4])
				      begin
				       $sformat(message, "%s lane1 %02x k= %01b", message, rxdata_int[4], rxdatak_int[4]);
				      end
				    if (rx_en_int[5])
				      begin
				       $sformat(message, "%s lane1 %02x k= %01b", message, rxdata_int[5], rxdatak_int[5]);
				      end
				    if (rx_en_int[6])
				      begin
				       $sformat(message, "%s lane1 %02x k= %01b", message, rxdata_int[6], rxdatak_int[6]);
				      end
				    if (rx_en_int[7])
				      begin
				       $sformat(message, "%s lane1 %02x k= %01b", message, rxdata_int[7], rxdatak_int[7]);
				      end				
				    $sformat(message, "%s at time\n", message, $time);      				      				      				      				      				      
				    $fwrite(LOGFILE, message);
				    check_identical = 1'b0;
				  end
				else
				  begin
				    check_identical = 1'b1;
				  end
  end
endfunction  

function packet_decode_dllp;
  input [7:0] packet_buf[4095:0];
  input integer packet_pointer;
  input integer time_val;
  integer i;
  string message;
  reg [7:0] masked_type;
  reg [7:0] pm_masked_type;
  reg flow_control_dllp;
  reg ack_nak_dllp;
  
  //DLLP parameters
  string dllp_type;
  reg vc_num;
  reg [11:0] seq_num;  
  reg [15:0] crc16;
  reg [7:0] hdrfc;
  reg [11:0] datafc;
  
  //Write entire packet to basic log file
 `ifdef DEBUG
  for (i = 0; i < packet_pointer; ++i)
  begin
    $sformat(message, "%02x", packet_buf[i]);
    $fwrite(LOGFILE, message);
  end
  $fwrite(LOGFILE, "\n\n");
 `endif
  
  //Start DLLP decode
  masked_type = (packet_buf[1] & 8'hF0);
  crc16 = {packet_buf[5], packet_buf[6]};  
  
  case (masked_type)
    `ACK:
    begin
      $sformat(dllp_type, "Ack");
      flow_control_dllp = 0;
      ack_nak_dllp = 1;
      //Ignore Reserved fields.  Should I print these?
      seq_num = {packet_buf[3][3:0], packet_buf[4]};
    end
    
    `NAK:
    begin
      $sformat(dllp_type, "Nak");    
      flow_control_dllp = 0;
      ack_nak_dllp = 1;
      //Ignore Reserved fields.  Should I print these?
      seq_num = {packet_buf[3][3:0], packet_buf[4]};
    end
    
    `PM_DLLP:
    begin
      pm_masked_type = packet_buf[1] & 8'h07;
      case (pm_masked_type)
      
      8'h00:
        begin
          $sformat(dllp_type, "PM_Enter_L1");
        end   
      
      8'h01:
        begin  
          $sformat(dllp_type, "PM_Enter_L23");
        end
      
      8'h03:
        begin
          $sformat(dllp_type, "PM_Active_state_Request_L1");
        end
        
      8'h04:
        begin
          $sformat(dllp_type, "PM_Request_Ack");
        end
      default:
        begin
          $sformat(dllp_type, "invalid DLLP");
        end
      endcase
      flow_control_dllp = 0;  
      ack_nak_dllp = 0;
    end
    
    `INIT_FC1_P:
    begin
      $sformat(dllp_type, "InitFC1-P");  
      flow_control_dllp = 1;   
      ack_nak_dllp = 0;   
    end
    
    `INIT_FC1_NP:
    begin
      $sformat(dllp_type, "InitFC1-NP");    
      flow_control_dllp = 1;    
      ack_nak_dllp = 0;    
    end
    
    `INIT_FC1_CPL:
    begin
      $sformat(dllp_type, "InitFC1-Cpl");  
      flow_control_dllp = 1;     
      ack_nak_dllp = 0;     
    end
    
    `INIT_FC2_P:
    begin
      $sformat(dllp_type, "InitFC2-P");   
      flow_control_dllp = 1;  
      ack_nak_dllp = 0;       
    end
    
    `INIT_FC2_NP:
    begin
      $sformat(dllp_type, "InitFC2-NP");  
      flow_control_dllp = 1;    
      ack_nak_dllp = 0;      
    end
    
    `INIT_FC2_CPL:
    begin
      $sformat(dllp_type, "InitFC2-Cpl"); 
      flow_control_dllp = 1; 
      ack_nak_dllp = 0;          
    end    
    
    `UPDATE_FC_P:
    begin
      $sformat(dllp_type, "UpdateFC-P");   
      flow_control_dllp = 1;    
      ack_nak_dllp = 0;     
    end

    `UPDATE_FC_NP:
    begin
      $sformat(dllp_type, "UpdateFC-NP");    
      flow_control_dllp = 1;     
      ack_nak_dllp = 0;   
    end    
    
    `UPDATE_FC_CPL:
    begin
      $sformat(dllp_type, "UpdateFC-Cpl"); 
      flow_control_dllp = 1;     
      ack_nak_dllp = 0;      
    end
    
    default:
    begin
        $sformat(dllp_type, "invlid DLLP");
        flow_control_dllp = 0;    
        ack_nak_dllp = 0;  
    end
  endcase
    
    if(flow_control_dllp)
      begin
        vc_num = packet_buf[1] & 8'h03;
        hdrfc = {packet_buf[2][5:0], packet_buf[3][7:6]};
        datafc = {packet_buf[3][3:0], packet_buf[4]};        
        $sformat(message, "DLLP,type,VC ID,HdrFC,DataFC,CRC16,time\n");
        $fwrite(LOGFILE_DLLP, message);
        $fwrite(LOGFILE, message);
        $sformat(message, "DLLP,%s,%02x,%02d,%04d,%04x,%10d\n", dllp_type, vc_num, hdrfc, datafc, crc16, time_val);
        $fwrite(LOGFILE_DLLP, message);
        $fwrite(LOGFILE, message);        
        packet_decode_dllp = 1;
      end
    else
      begin
        if(ack_nak_dllp)
          begin
            $sformat(message, "DLLP,type,seq,CRC16,time\n");
            $fwrite(LOGFILE_DLLP, message);
            $fwrite(LOGFILE, message);
            $sformat(message, "DLLP,%s,%04d,%04x,%10d\n", dllp_type, seq_num, crc16, time_val);
            $fwrite(LOGFILE_DLLP, message);
            $fwrite(LOGFILE, message);          
            $fwrite(LOGFILE, "time is %d\n", time_val);  
            packet_decode_dllp = 1;
          end
        else
          begin
            $sformat(message, "DLLP,type,CRC16,time\n");
            $fwrite(LOGFILE_DLLP, message);
            $fwrite(LOGFILE, message);
            $sformat(message, "DLLP,%s,%04x,%10d\n", dllp_type, crc16, time_val);
            $fwrite(LOGFILE_DLLP, message);
            $fwrite(LOGFILE, message);       
            $fwrite(LOGFILE, "time is %d\n", time_val);     
            packet_decode_dllp = 1;            
          end
      end
    
    $fwrite(LOGFILE_DLLP, "\n");
	$fwrite(LOGFILE, "\n");
    
    
  
endfunction

function packet_decode;
  input [7:0] packet_buf[4095:0];
  input integer packet_pointer;
  input integer time_val;
  
  integer i;
  string message;
  reg [2:0] fmt;  
  reg [4:0] pkt_type;
  string type_str;
  string type_str_det;
  integer valid_transaction;
  reg [2:0] tc_field;
  reg [2:0] attr_field;
  reg th_field;
  reg td_field;
  reg ep_field;
  reg [1:0] at_field;
  reg [9:0] length_field;
  reg [11:0] sequence_field;
  reg [31:0] lcrc;
  reg		  dw3_ndw4;
  reg [63:0] address;
  integer	payload_pointer;
  reg [7:0]	bus_num;
  reg [4:0] dev_num;
  reg [2:0] func_num;
  reg [7:0] byte_en;
  reg [15:0] req_id;
  reg [7:0] tag_id;
  
  //Initialise
  valid_transaction = 0;
  
  //Start TLP decode, byte 0 is start TLP, so ignore that
  
  //Link Layer Fields
  sequence_field[11:8] = packet_buf[1][3:0];
  sequence_field[7:0] = packet_buf[2];
  lcrc[7:0] = packet_buf[packet_pointer - 2];
  lcrc[15:8] = packet_buf[packet_pointer - 3];
  lcrc[23:16] = packet_buf[packet_pointer - 4];
  lcrc[31:24] = packet_buf[packet_pointer - 5];
  
  //TLP decode
  fmt = packet_buf[3][7:5];
  pkt_type = packet_buf[3][4:0];
  case (fmt)
    `DW3_NODAT:
    begin
      case(pkt_type)
        `MRD:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Mem"); 
		  $sformat(type_str_det, "MRd(32)"); 
        end
        `MRDLK:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Mem"); 
		  $sformat(type_str_det, "MRdLk(32)"); 
        end
        `IORD:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "IO"); 
		  $sformat(type_str_det, "IORd"); 
        end
        `CFGRD0:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cfg"); 
		  $sformat(type_str_det, "CfgRd0");
        end
        `CFGRD1:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cfg"); 
		  $sformat(type_str_det, "CfgRd1"); 
        end
        `CPL:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cpl"); 
		  $sformat(type_str_det, "Cpl");
        end
        `CPLLK:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cpl"); 
		  $sformat(type_str_det, "CplLk"); 
        end
        default:
        begin
          $fwrite(LOGFILE_TLP,"unsupported 3DW header type\n");
		  $fwrite(LOGFILE,"unsupported 3DW header type\n");
        end
      endcase 
    end
    `DW4_NODAT:
    begin
      if(pkt_type[4] == 1'b0) //Not message
        begin
          case(pkt_type)
          `MRD:
            begin
              valid_transaction = 1;
			  $sformat(type_str, "Mem"); 
			  $sformat(type_str_det, "MRd(64)"); 
            end
            `MRDLK:
            begin
              valid_transaction = 1;
			  $sformat(type_str, "Mem"); 
			  $sformat(type_str_det, "MRdLk(64)"); 
            end
            `CPL:
            begin
              valid_transaction = 1;
			  $sformat(type_str, "Cpl"); 
			  $sformat(type_str_det, "Cpl(64)"); 
            end
            `CPLLK:
            begin
              valid_transaction = 1;
			  $sformat(type_str, "Cpl"); 
			  $sformat(type_str_det, "CplLk(64)"); 
            end
            default:
            begin
              $fwrite(LOGFILE_TLP,"unsupported 4DW header type\n");
			  $fwrite(LOGFILE,"unsupported 4DW header type\n");
            end            
          endcase 
        end
      else   
      begin  
        $fwrite(LOGFILE, "TLP, message");
      end
    end
    `DW3_DAT:
    begin
      case (pkt_type)
        `MWR :
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Mem");
		  $sformat(type_str_det, "MWr(32)");
        end
        `IOWR:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "IO");
		  $sformat(type_str_det, "IOWr");
        end
        `CFGWR0:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cfg");
		  $sformat(type_str_det, "CfgWr0");
        end
        `CFGWR1:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cfg");
		  $sformat(type_str_det, "CfgWr1");
        end
        `CPLD:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cpl");
		  $sformat(type_str_det, "CplD");
        end
        `CPLDLK:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Cpl");
		  $sformat(type_str_det, "CplLkD");
        end
        `FETCHADD:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Atom");
		  $sformat(type_str_det, "FetAdd(32)");
        end
        `SWAP:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Atom");
		  $sformat(type_str_det, "Swap(32)");
        end
        `CAS:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Atom");
		  $sformat(type_str_det, "CAS(32)");
        end
        default:
        begin
          $fwrite(LOGFILE_TLP,"unsupported 4DW header type\n");
		  $fwrite(LOGFILE,"unsupported 4DW header type\n");
        end
      endcase
    end
    `DW4_DAT:
    begin
      case (pkt_type)
        `MWR :
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Mem");
		  $sformat(type_str_det, "MemWr(64)");
        end
        `FETCHADD:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Atom");
		  $sformat(type_str_det, "FetAdd(64)");
        end
        `SWAP:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Atom");
		  $sformat(type_str_det, "Swap(32)");
        end
        `CAS:
        begin
          valid_transaction = 1;
		  $sformat(type_str, "Atom");
		  $sformat(type_str_det, "CAS(32)");
        end
        default:
        begin
          $fwrite(LOGFILE_TLP,"unsupported 4DW header type\n");
		  $fwrite(LOGFILE,"unsupported 4DW header type\n");
        end
      endcase
    end
    `TLP_PRE:
    begin
      $fwrite(LOGFILE, "TLP, unsupported fmt field (TLP support TBD)\n");      
	  $fwrite(LOGFILE_TLP, "TLP, unsupported fmt field (TLP support TBD)\n");    
    end
    default:
    begin
      $fwrite(LOGFILE, "TLP, unsupported fmt field\n");
	  $fwrite(LOGFILE_TLP, "TLP, unsupported fmt field\n");
    end
  endcase
  
    //Function decoding
  if (pkt_type == (`CFGRD0 | `CFGWR0 | `CFGRD1 | `CFGWR1))
	begin
		bus_num = packet_buf[11];
		dev_num = packet_buf[12][7:3];
		func_num = packet_buf[12][2:0];
	end
  
  //Byte enable decoding
  if (pkt_type == (`CFGRD0 | `CFGWR0 | `CFGRD1 | `CFGWR1 | `MWR | `MRD | `MRDLK | `IORD | `IOWR | `FETCHADD | `SWAP | `CAS))
	begin
		byte_en = packet_buf[10];
	end
  
  //Address decoding
  if (pkt_type == (`MWR | `MRD | `MRDLK | `IORD | `IOWR | `FETCHADD | `SWAP | `CAS))
  begin
	if (fmt[0]) //64bit
	begin
		address[63:56] = packet_buf[11];
		address[55:48] = packet_buf[12];
		address[47:40] = packet_buf[13];
		address[39:32] = packet_buf[14];
		address[31:24] = packet_buf[15];
		address[23:16] = packet_buf[16];
		address[15:8] = packet_buf[17];
		address[7:0] = packet_buf[18];		
		payload_pointer = 19;		
	end
	else
	begin
		address[63:32] = 0;
		address[31:24] = packet_buf[11];
		address[23:16] = packet_buf[12];
		address[15:8] = packet_buf[13];
		address[7:0] = packet_buf[14];
		payload_pointer = 15;
	end
end

	//BE and transaction ID 
	if (pkt_type == (`MWR | `MRD | `MRDLK | `IORD | `IOWR | `FETCHADD | `SWAP | `CAS))
	begin
		req_id[15:8] = packet_buf[7];
		req_id[7:0] = packet_buf[8];
		tag_id = packet_buf[9];
		
	end
    
  if(valid_transaction)
    begin
      tc_field = packet_buf[4][6:4];
      attr_field[2] = packet_buf[4][2];
      attr_field[1:0] = packet_buf[5][3:2];
      th_field = packet_buf[4][0];
      td_field = packet_buf[5][7];
      ep_field = packet_buf[5][6];
      at_field = packet_buf[5][3:2];
      length_field[9:8] = packet_buf[5][1:0];
      length_field[7:0] = packet_buf[6];
	  
	  $sformat(message, "TLP,%s,%s,TC,TH,TD,EP,Attr,AT,Len,LCRC,Time", type_str, type_str_det);
	  if (pkt_type == (`CFGRD0 | `CFGWR0 | `CFGRD1 | `CFGWR1))
	  begin
		$sformat(message, "%s, bus, device, function", message);
	  end
	  /*if (pkt_type == (`MWR | `MRD | `MRDLK | `IORD | `IOWR | `FETCHADD | `SWAP | `CAS))
	  begin
		$sformat(message, "%*/
	  $sformat(message, "%s\n", message);
      $fwrite(LOGFILE, message);
	  
	  $sformat(message, "%s,%s,TC,TH,TD,EP,Attr,AT,Len,Time", type_str, type_str_det);
	  if (pkt_type == (`CFGRD0 | `CFGWR0 | `CFGRD1 | `CFGWR1))
	  begin
		$sformat(message, "$s, bus, device, function", message);
	  end
	  $sformat(message, "%s\n", message);
	  $fwrite(LOGFILE_TLP, message);	
	  
      $sformat(message, "%04d,%s,%s,%01x,%01x,%01x,%01x,%01x,%01x,%d,%08x,%10d", sequence_field, type_str, type_str_det, tc_field, th_field, td_field, ep_field, attr_field, at_field, length_field, lcrc,time_val);
	  if (pkt_type == (`CFGRD0 | `CFGWR0 | `CFGRD1 | `CFGWR1))
	  begin
		$sformat(message, "$s, %02x, %02x, %02x", message, bus_num, dev_num, func_num);
	  end
	  $sformat(message, "%s\n", message);
      $fwrite(LOGFILE, message);
	  
	  $sformat(message, "%s,%s,%01x,%01x,%01x,%01x,%01x,%01x,%d,%10d\n", type_str, type_str_det, tc_field, th_field, td_field, ep_field, attr_field, at_field, length_field, time_val);
	  if (pkt_type == (`CFGRD0 | `CFGWR0 | `CFGRD1 | `CFGWR1))
	  begin
		$sformat(message, "$s, %02x, %02x, %02x", message, bus_num, dev_num, func_num);
	  end
	  $sformat(message, "%s\n", message);
	  $fwrite(LOGFILE_TLP, message);
	  $fwrite(LOGFILE, "\n");
	  $fwrite(LOGFILE_TLP, "\n");
    end    
	
	//Length check
	if(td_field)
	begin
		if (dw3_ndw4)
		begin
			if (length_field - (packet_pointer - 5 - 12 - 2 - 4))
			begin
				$fwrite(LOGFILE, "!!incorrect length TLP!!\n");
			end
		end
		else
		begin
			if (length_field - (packet_pointer - 5 - 16 - 2 - 4))
			begin
				$fwrite(LOGFILE, "!!incorrect length TLP!!\n");
			end
		end
	end
	else
	begin
		if (dw3_ndw4)
		begin
			if (length_field - (packet_pointer - 5 - 12 - 2))
			begin
				$fwrite(LOGFILE, "!!!incorrect length TLP!!!\n");
			end
		end
		else
		begin
			if (length_field - (packet_pointer - 5 - 16 - 2))
			begin
				$fwrite(LOGFILE, "!!!incorrect length TLP!!!\n");
			end
		end
	end
	
    //Write entire packet to basic log file
	$fwrite(LOGFILE, "full packet:\n");
	$fwrite(LOGFILE_TLP, "full packet:\n");
  for (i = 0; i < packet_pointer; ++i)
  begin
    $sformat(message, "%02x", packet_buf[i]);
    $fwrite(LOGFILE, message);
	$fwrite(LOGFILE_TLP, message);
  end
  $fwrite(LOGFILE, "\n\n");	
  $fwrite(LOGFILE_TLP, "\n\n");
  
  packet_decode = 1;
endfunction
	
endmodule